/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIALYSTOK 2018                               **
**                                                                        **
**              Przyklad obslugi protokolu sieciowego: ARP                **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2018.03.09                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <LPC23xx.H>
#include "system.h"
#include "types.h"
#include "irqvectors.h"
#include "serialconst.h"
#include "uart0.h"
#include "strings.h"
#include "emacdriver.h"
#include "ipmemservice.h"
#include "frameutil.h"

#define IRQCallCntLimit                 500

#define StandardLifeTime         1280

#define ARP_request              1
#define ARP_response             2
#define RARP_request             3
#define RARP_response            4

typedef struct {
                 USHORT               HardwareType ;                           /* =  2 */
                 USHORT               ProtocolType ;                           /* =  2 */
                 USHORT               AddressLength ;                          /* =  2 */
                 USHORT               Operation ;                              /* =  2 */
                 UCHAR                SenderMACAddress [ EthMACAddressSize ] ; /* =  6 */
                 ULONG                SenderIPAddress ;                        /* =  4 */
                 UCHAR                TargetMACAddress [ EthMACAddressSize ] ; /* =  6 */
                 ULONG                TargetIPAddress ;                        /* =  4 */
               } ARPHeaderRecT ;                                               /* = 28 */
#define HardwareTypeOffset            0
#define ProtocolTypeOffset            HardwareTypeOffset + 2
#define AddressLengthOffset           ProtocolTypeOffset + 2
#define OperationOffset               AddressLengthOffset + 2
#define SenderMACAddressOffset        OperationOffset + 2
#define SenderIPAddressOffset         SenderMACAddressOffset + EthMACAddressSize
#define TargetMACAddressOffset        SenderIPAddressOffset + 4
#define TargerIPAddressOffset         TargetMACAddressOffset + EthMACAddressSize
#define ARPHeaderRecTSize             28



static UCHAR Broadcast [ 6 ] = { 0xFF , 0xFF , 0xFF , 0xFF , 0xFF , 0xFF } ;

#define BufferSize               64
               
typedef struct {
                 EMACInstanceRecT *          EMACLayer ;
                 UCHAR                       Buffer [ BufferSize + 1 ] ;
                 USHORT                      Index ;
               } ARPDemoInstanceRecT ;
typedef ARPDemoInstanceRecT * PtrARPDemoInstanceRecT ;

static ARPDemoInstanceRecT ARPDemoInstance ;
static ULONG IRQCallCounter ;

#define MachineIPAddress        0xC0A80037
//                              ^^^ 192.168.0.55

UCHAR MyMACAddr [ 6 ]           = { 0x08 , 0x14 , 0x16 , 0x22 , 0x28 , 0x2D } ;


static void SoftwareInit ( void )
{
  /*-------------------------------------------------------------------------*/
  IRQCallCounter = 0 ;
  ARPDemoInstance . EMACLayer = NIL ;
  ARPDemoInstance . Index = 0 ;
  ARPDemoInstance . Buffer [ 0 ] = 0 ;
} /* SoftwareInit */


static __irq void T1_IRQHandler ( void )
{
  /*-------------------------------------------------------------------------*/
  IRQCallCounter ++ ;
  if ( IRQCallCounter >= IRQCallCntLimit )
  {
    IRQCallCounter = 0 ;
  } /* if */ ;
  T1IR = 1 ;
  VICVectAddr = 0 ;
} /* T1_IRQHandler */


static void HardwareInit ( void ) 
{
  /*-------------------------------------------------------------------------*/
  PINSEL10 = 0 ;
  T1MR0 = ( PCLK / 1000 ) - 1 ;
  T1MCR = 3 ;
  T1TCR = 1 ;
  VICVectAddr5 = ( unsigned long ) T1_IRQHandler ;
  VICVectCntl5 = 15 ;
  VICIntEnable = ( 1 << TIMER1_Interrupt ) ;
  UART0InitEnvir ( ) ;
} /* HardwareInit */


static void DisplayMAC ( UCHAR * MACAddr )
{  
  UCHAR NumberStr [ 32 ] ;
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < 5 ; Loop ++ )
  {
    WordToHexStr ( NumberStr , 2 , '0' ,  ( USHORT ) ( * MACAddr ) ) ;
    MACAddr ++ ;
    UART0SendString ( NumberStr ) ;
    UART0Send ( '-' ) ;
  } /* for */ ;  
  WordToHexStr ( NumberStr , 2 , '0' ,  ( USHORT ) ( * MACAddr ) ) ;
  UART0SendString ( NumberStr ) ;
} /* DisplayMAC */


static void SendARPResponse ( EMACInstanceRecT * InterfInstance ,
                              ARPHeaderRecT *    ARPPacked )
{
  PtrOSFrameRecT OutFrame ;
  /*-------------------------------------------------------------------------*/
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( OutFrame )
  {
    OutFrame -> Index = 0 ;
    StartEthFrame ( OutFrame , ARPPacked -> SenderMACAddress , InterfInstance -> InterfaceMACAddress ,
                    FRAME_ARP ) ;
    AddUSHORTSwapField ( OutFrame , EthHardwType ) ;
    AddUSHORTSwapField ( OutFrame , FRAME_IP ) ;
    AddUSHORTSwapField ( OutFrame , IP_HLen_PLen ) ;
    AddUSHORTSwapField ( OutFrame , ARP_response ) ;
    AddByteArrayField ( OutFrame , InterfInstance -> InterfaceMACAddress , EthMACAddressSize ) ;
    AddULONGSwapField ( OutFrame , MachineIPAddress ) ;
    AddByteArrayField ( OutFrame , ARPPacked -> SenderMACAddress , EthMACAddressSize ) ;
    AddULONGSwapField ( OutFrame , ARPPacked -> SenderIPAddress ) ;
    InterfInstance -> InterfaceSendFrame ( OutFrame ) ;
    InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  } /* if */ ;
} /* SendARPResponse */

UCHAR ARPMessage1 [ ]        = "\r\nOtrzymalismy odpowiedz na zapytanie ARP\r\nMaszyna o IP=" ;
UCHAR ARPMessage2 [ ]        = " ma MAC=" ;
UCHAR ARPMessage3 [ ]        = "\r\n" ;

static void RecvARPResponse ( EMACInstanceRecT * InterfInstance ,
                              ARPHeaderRecT *    ARPPacked )
{
  UCHAR NumberStr [ 20 ] ;
  /*-------------------------------------------------------------------------*/
  UART0SendString ( ARPMessage1 ) ;
  IPToString ( NumberStr , ARPPacked -> SenderIPAddress ) ;
  UART0SendString ( NumberStr ) ;
  UART0SendString ( ARPMessage2 ) ;
  DisplayMAC ( ARPPacked -> SenderMACAddress ) ;
  UART0SendString ( ARPMessage3 ) ;
} /* RecvARPResponse */


static void ProcessARPPacked ( PtrOSFrameRecT Frame )
{
  ARPHeaderRecT ARPPacked ;
  EtherPackedHeaderRecT EtherPackedHeader ;  
  /*-------------------------------------------------------------------------*/
  ARPPacked . HardwareType = GetUSHORTSwapField ( Frame , EtherPackedHeaderRecTSize + HardwareTypeOffset ) ;
  ARPPacked . ProtocolType = GetUSHORTSwapField ( Frame , EtherPackedHeaderRecTSize + ProtocolTypeOffset ) ;
  ARPPacked . AddressLength = GetUSHORTSwapField ( Frame , EtherPackedHeaderRecTSize + AddressLengthOffset ) ;
  ARPPacked . Operation = GetUSHORTSwapField ( Frame , EtherPackedHeaderRecTSize + OperationOffset ) ;
  GetByteArrayField ( ARPPacked . SenderMACAddress , Frame , EtherPackedHeaderRecTSize + SenderMACAddressOffset ,
                      EthMACAddressSize ) ;
  ARPPacked . SenderIPAddress = GetULONGSwapField ( Frame , EtherPackedHeaderRecTSize + SenderIPAddressOffset ) ;
  GetByteArrayField ( ARPPacked . TargetMACAddress , Frame , EtherPackedHeaderRecTSize + TargetMACAddressOffset ,
                      EthMACAddressSize ) ;
  ARPPacked . TargetIPAddress = GetULONGSwapField ( Frame , EtherPackedHeaderRecTSize + TargerIPAddressOffset ) ;
  if ( ARPPacked . HardwareType == EthHardwType ) 
  {
    if ( ARPPacked . ProtocolType == FRAME_IP )
    {
      if ( ARPPacked . AddressLength == IP_HLen_PLen )
      {
        switch ( ARPPacked . Operation )
        {
          case ARP_request              :
            GetEthHeader ( & EtherPackedHeader , Frame ) ;
            if ( ARPPacked . TargetIPAddress == MachineIPAddress )
            {
              SendARPResponse ( ARPDemoInstance . EMACLayer , & ARPPacked ) ;
            } /* if */ ;
            break ;
          case ARP_response             :
            RecvARPResponse ( ARPDemoInstance . EMACLayer , & ARPPacked ) ;
            break ;
          case RARP_request             :
            break ;
          case RARP_response            :
            break ;
          default                       :
            break ;
        } /* switch */ ;
      } /* if */ ;
    } /* if */ ;
  } /* if */ ;
} /* ProcessARPPacked */


static void FrameProcessing ( PtrOSFrameRecT Frame )
{
  EtherPackedHeaderRecT EtherPackedHeader ;  
  /*-------------------------------------------------------------------------*/
  GetEthHeader ( & EtherPackedHeader , Frame ) ;
  SetIPHeaderOrgin ( Frame ) ;
  if ( EtherPackedHeader . LengthOrType >= 1518 )
  {
    if ( EtherPackedHeader . LengthOrType == FRAME_ARP )
    {
      ProcessARPPacked ( Frame ) ;
    } /* if */ ;
  } /* if */ ;
  ARPDemoInstance . EMACLayer -> InterfaceDeallocFrameMemory ( Frame ) ;
} /* FrameProcessing */


UCHAR ARPMessage4 [ ]        = "\r\nWysylamy zapytanie o MAC do maszyny o IP=" ;
UCHAR ARPMessage5 [ ]        = "\r\n" ;

void PrepareARPRequest ( EMACInstanceRecT * InterfInstance ,
                         ULONG              DestIPAddress )
{
  PtrOSFrameRecT OutFrame ;
  UCHAR DestMACAddress [ EthMACAddressSize ] ;
  ULONG Loop ;
  UCHAR NumberStr [ 20 ] ;
  /*-------------------------------------------------------------------------*/
  UART0SendString ( ARPMessage4 ) ;
  IPToString ( NumberStr , DestIPAddress ) ;
  UART0SendString ( NumberStr ) ;
  UART0SendString ( ARPMessage5 ) ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( OutFrame )
  {
    OutFrame -> Index = 0 ;
    for ( Loop = 0 ; Loop < EthMACAddressSize ; Loop ++ )
      DestMACAddress [ Loop ] = 0 ;
    StartEthFrame ( OutFrame , Broadcast , InterfInstance -> InterfaceMACAddress ,
                    FRAME_ARP ) ;
    AddUSHORTSwapField ( OutFrame , EthHardwType ) ;
    AddUSHORTSwapField ( OutFrame , FRAME_IP ) ;
    AddUSHORTSwapField ( OutFrame , IP_HLen_PLen ) ;
    AddUSHORTSwapField ( OutFrame , ARP_request ) ;
    AddByteArrayField ( OutFrame , InterfInstance -> InterfaceMACAddress , EthMACAddressSize ) ;
    AddULONGSwapField ( OutFrame , MachineIPAddress ) ;
    AddByteArrayField ( OutFrame , DestMACAddress , EthMACAddressSize ) ;
    AddULONGSwapField ( OutFrame , DestIPAddress ) ;
    InterfInstance -> InterfaceSendFrame ( OutFrame ) ;
    InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  } /* if */ ;
} /* PrepareARPRequest */

UCHAR TraceInput1 [ ]       = "\r\nWczytano: " ;
UCHAR TraceInput2 [ ]       = "\r\n" ;

static void SerialPool ( void )
{
  UCHAR Ch ;
  ULONG DersinationIP ;
  /*-------------------------------------------------------------------------*/
  if ( UART0DataPresent ( ) )
  {
    Ch = UART0GetData ( ) ;
    if ( Ch == 0x0D )
    {
      UART0SendString ( TraceInput1 ) ;
      UART0SendString ( ARPDemoInstance . Buffer ) ;
      UART0SendString ( TraceInput2 ) ;     
      if ( StrToIP ( ARPDemoInstance . Buffer , & DersinationIP ) )
      {
        PrepareARPRequest ( ARPDemoInstance . EMACLayer , DersinationIP ) ;
      } /* if */ ;
      ARPDemoInstance . Index = 0 ;
      ARPDemoInstance . Buffer [ 0 ] = 0 ;
      return ;
    } /* if */ ;
    if ( ARPDemoInstance . Index < BufferSize )
    {
      ARPDemoInstance . Buffer [ ARPDemoInstance . Index ] = Ch ;\
      ARPDemoInstance . Index ++ ;
      ARPDemoInstance . Buffer [ ARPDemoInstance . Index ] = 0 ;
    } /* if */ ;
  } /* if */ ;
} /* SerialPool */


static void StartNet ( PtrARPDemoInstanceRecT Instance )
{
  EtherSpeedMode EthMode ;
  UCHAR LocalMyMACAddr [ 6 ] ;
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < 6 ; Loop ++ )
    LocalMyMACAddr [ Loop ] = MyMACAddr [ Loop ] ;
  EthMode = Mode_AutoNeg ;
//  EthMode = Mode_10MBIT ;
//  EthMode = Mode_100MBIT ;
  Instance -> EMACLayer = InitPHYLayerInstance ( LocalMyMACAddr , EthMode ) ;
  InitIPLayerMemoryService ( ) ;
  ARPDemoInstance . EMACLayer -> InterfaceProcessFrame = FrameProcessing ;
  ARPDemoInstance . EMACLayer -> InterfaceAllocFrameMemory = FrameMemoryAllocate ;
  ARPDemoInstance . EMACLayer -> InterfaceDeallocFrameMemory = FrameMemoryDeallocate ;
} /* StartNet */


UCHAR Hello1 [ ]      = "\r\nHello: obsluga ARP\r\n" ;
UCHAR Hello2 [ ]      = "\r\nMoj adres IP=" ;
UCHAR Hello5 [ ]      = "\r\nAdres MAC=" ;
UCHAR Hello6 [ ]      = "\r\n************************\r\n" ;


static void HelloMessage ( void )
{
  UCHAR NumberStr [ 20 ] ;
  /*-------------------------------------------------------------------------*/
  UART0SendString ( Hello1 ) ;
  UART0SendString ( Hello2 ) ;
  IPToString ( NumberStr , MachineIPAddress ) ;
  UART0SendString ( NumberStr ) ;
  UART0SendString ( Hello5 ) ;
  DisplayMAC ( MyMACAddr ) ;
  UART0SendString ( Hello6 ) ;
} /* HelloMessage */


int main ( void )
{
  /*-------------------------------------------------------------------------*/
  SysInit ( ) ;
  HardwareInit ( )  ;
  SoftwareInit ( ) ;
  UART0HardwInit( FOSC , N81Mode , 9600 ) ;
  HelloMessage ( ) ;
  StartNet ( & ARPDemoInstance ) ;
  for ( ; ; )
  {
    ARPDemoInstance . EMACLayer -> CheckEthInterface ( ) ;
    SerialPool ( ) ;
  } /* loop */ ;
} /* main */

